<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

require_once (__DIR__ . '/word.php');


/* This class feeds data about documents into a search engine database. The
 * class does not support altering data about already indexed documents. The
 * intended scenario is that the database will be rebuild by a web crawler
 * from scratch regularly to be up-to-date.
 */
final class feeder {

// The maximum number of rows inserted into the database with a single
// INSERT OR IGNORE statement must not be bigger than the default value of 500
// for SQLITE_MAX_COMPOUND_SELECT in SQLite.
private const max_compound_insert = 496;


/* Property
 */

// SQLite3 object used to connect to the actual database
private object $db;


/* Methods
 */

/* The constructor connects to the SQLite database file located at $path. Using
 * journal_mode WAL improves concurrency, see https://www.sqlite.org/wal.html .
 */
public function __construct (
	string $path
)
{
	$this->db = new \SQLite3 ($path, SQLITE3_OPEN_READWRITE);
	$this->db->busyTimeout (10000); // 10 seconds
	$this->db->exec ('PRAGMA journal_mode=WAL;');
}


/* The destructor closes the connection to the database, so we do not do that
 * explicitely anywhere else.
 */
public function __destruct (
)
{
	$this->db->close ();
}


/* Creates a new SQLite3 database, whose filename is provided by $path, and
 * creates the tables needed for the search engine inside it. If the database
 * file already exists beforehand, it is deleted to start anew. DB description:
 *
 * /<document>/
 *  <id>     auto-increment
 *  <url>    URL of the document
 *  <size>   number of words in the content
 *  <name>   e.g. HTML <title> element content or filename of a text file
 *  <desc>   description, e.g. HTML <meta name=description> or a lead paragrapgh
 *  <author> the creator of the document's main content
 *  <flags>  flags reserved for later use, e.g. availability of a thumbnail
 *
 * /<title>/
 *  <word> a word which occurs somewhere in the titles of webpages
 *  <docs> a list of all occurrences of the word in all webpage titles
 * 
 * /<content>/
 *  <word> a word which occurs somewhere in the content of webpages
 *  <docs> a list of all occurrences of the word in all webpage contents
 *
 * The format of docs is like "id:dist,dist,dist,dist.id:dist.id:dist,dist."
 * where id is a document id and distance is how many words from the start of
 * the title/content or from the last occurrence the word in question occurred.
 */
public static function create (
	string $path
):	void
{
	file_exists ($path) and unlink ($path);
	$db = new \SQLite3 ($path, SQLITE3_OPEN_CREATE | SQLITE3_OPEN_READWRITE);
	$db->exec ("CREATE TABLE document (id INTEGER PRIMARY KEY ASC AUTOINCREMENT, url TEXT UNIQUE NOT NULL, size INTEGER NOT NULL, name TEXT DEFAULT '', desc TEXT DEFAULT '', author TEXT DEFAULT '', flags INTEGER DEFAULT 0)");
	$db->exec ("CREATE TABLE title (word TEXT PRIMARY KEY ASC, docs TEXT DEFAULT '') WITHOUT ROWID");
	$db->exec ("CREATE TABLE content (word TEXT PRIMARY KEY ASC, docs TEXT DEFAULT '') WITHOUT ROWID");
	$db->close ();
}


/* Takes an array of integers and turns it into a comma-separated string of the
 * corresponding differences: the first number in the output is the same as the
 * first number in the input, the second number in the output is the difference
 * of the second and first number in the input, the third number in the output
 * is the difference of the third and the second number in the input ... In
 * other words, if we interpret the input array as partial sums of a series, the
 * output string contains the underlying sequence. The reason for computing the
 * differences is that they encode the same information with less memory.
 * Example: distances ([995, 1033, 1034]) => "995,38,1".
 */
private static function distances (
	array $partialsums
):	string
{
	$sequence = [];
	$previous = 0;
	foreach ($partialsums as $partialsum):
		$sequence[] = $partialsum - $previous;
		$previous = $partialsum;
	endforeach;
	return implode (',', $sequence);
}


/* Returns the highest document id from the database. This is also the id of the
 * newest document registered in the database and reperesents the number of
 * documents in the database, since the id is an autoincrement field and we
 * never delete any documents/rows from the database.
 */
private function maxid (
):	int
{
	return (int) $this->db->querySingle ("SELECT MAX(id) FROM document");
}


/* Registers and indexes a document in the database. Charset UTF-8 is expected
 * for its $url, its $title and $content.
 */
public function write (
	string $url,
	string $title,
	string $content,
	string $description,
	string $creator,
	int $flags = 0
):	void
{
	// prepare title and content for indexing
	$name = word::utf8 ($title);
	$desc = word::utf8 ($description);
	$author = word::utf8 ($creator);
	$t_index = word::occurrences ($name);
	$c_index = word::occurrences ($content);

	// count the words in the content; this is faster than
	// array_sum (array_map ('count', $c_index))
	$size = 0;
	foreach ($c_index as $places):
		$size += count ($places);
	endforeach;

	// register the document
	$insert = $this->db->prepare ("INSERT INTO document (url, size, name, desc, author, flags) VALUES (?, ?, ?, ?, ?, ?)");
	$insert->bindValue (1, $url, SQLITE3_TEXT);
	$insert->bindValue (2, $size, SQLITE3_INTEGER);
	$insert->bindValue (3, $name, SQLITE3_TEXT);
	$insert->bindValue (4, $desc, SQLITE3_TEXT);
	$insert->bindValue (5, $author, SQLITE3_TEXT);
	$insert->bindValue (6, $flags, SQLITE3_INTEGER);
	$insert->execute ();
	$insert->close ();

	// try hard to get the document id
	$id = (int) $this->db->lastInsertRowID () ?: $this->maxid ();

	// enter data
	if ($id):
		$this->write_to_table ('title', $id, $t_index);
		$this->write_to_table ('content', $id, $c_index);
	else:
		trigger_error (__METHOD__ . " could not get document id for '$url'");
	endif;
}


/* Saves for each word of either the 'title' or 'content' (determined by $table)
 * of a document identified by $id at which places it occurred. $index must be
 * an array whose keys are all distinct words in the text and the values are
 * arrays of integers which give the places (1 for the first word, 2 for the
 * second word, ...) of the respective word in the text.
 */
private function write_to_table (
	string $table, // must be 'title' or 'content'
	int $id,
	array $index
):	void
{
	if (empty ($index)):
		return;
	endif;

	// register all words in the database that are unknown so far
	$words = [];
	foreach ($index as $word => $places):
		$words[] = \SQLite3::escapeString ((string) $word);
	endforeach;
	foreach (array_chunk ($words, self::max_compound_insert) as $chunk):
		$this->db->exec ("INSERT OR IGNORE INTO $table (word) VALUES ('" . implode ("'), ('", $chunk) . "')");
	endforeach;
	unset ($words);

	// update the words in the database with their places in this document
	$this->db->exec ('BEGIN TRANSACTION');
	$update = $this->db->prepare ("UPDATE $table SET docs = docs || $id || ':' || ? || '.' WHERE word = ?");
	$update->bindParam (1, $loci, SQLITE3_TEXT);
	$update->bindParam (2, $word, SQLITE3_TEXT);
	foreach ($index as $word => $places):
		$loci = self::distances ($places);
		$update->execute ();
		$update->reset ();
	endforeach;
	$update->close ();
	$this->db->exec ('COMMIT TRANSACTION');
}

} // class feeder

// The End