<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// declare (strict_types = 1); // only during development

namespace prlbr\mouse;

/* Converts characters with special function in HTML to HTML entities.
 */
function esc (
	string $string
):	string
{
	return htmlspecialchars ($string, ENT_SUBSTITUTE | ENT_QUOTES, 'UTF-8');
}


/* Returns information useful for setting up a cronjob and subscribing to the
 * newsfeed, formatted as HTML.
 */
function help (
	string $module_url,
	string $home_url,
	string $id
):	string
{
	if (
		isset ($_SERVER['REQUEST_SCHEME'])
		and isset ($_SERVER['HTTP_HOST'])
		and $_SERVER['REQUEST_SCHEME']
		and $_SERVER['HTTP_HOST']
	):
		$base = esc ($_SERVER['REQUEST_SCHEME'] . '://' . $_SERVER['HTTP_HOST']);
	else:
		$base = '';
	endif;

	// Aneamal root relative to the Document root:
	$home = esc ($home_url);

	// Crawler address relative to the Aneamal root:
	$script = esc (substr ($module_url, strlen ($home_url)) . '/crawl.php?' . $id);

	// Atom feed address relative to the Aneamal root:
	$atom = esc ("/aneamal/public/mouse/$id.atom");

	// TODO: change this when the version changes.
	$version = 'Mouse/2';

	return <<<MUS
		<div id='mouse' class='mouse-results'>
		<dl>
		<dt>Crawler’s URL</dt>
		<dd><samp><a href='$base$home$script'>$base$home$script</a></samp></dd>
		<dt>Trigger in Aneamal</dt>
		<dd><samp>@javascript: -&gt;$script</samp></dd>
		<dt>Atom Feed</dt>
		<dd><samp><a href='$base$home$atom'>$base$home$atom</a></samp></dd>
		<dt>Atom in Aneamal</dt>
		<dd><samp>@atom: -&gt;$atom</samp></dd>
		<dt>Version</dt>
		<dd>$version</dd>
		</dl>
		</div>
		MUS;
}


/* Returns a single search result item formatted as HTML.
 */
function item (
	array $result,
	int $snippetlen = 0,
	string $attributes = ''
):	string
{
	if (empty ($result)):
		return '';
	endif;

	if ($attributes !== '' and $attributes[0] !== "\x20"):
		$attributes = "\x20" . $attributes;
	endif;

	$url = esc ($result['url']);
	$title = $result['name'] === ''? '???': esc ($result['name']);

	if (isset ($result['desc']) and $result['desc'] !== '' and $snippetlen > 0):
		if (mb_strlen ($result['desc'], 'UTF-8') > $snippetlen):
			$result['desc'] = mb_substr ($result['desc'], 0, $snippetlen);
			if (mb_substr ($result['desc'], -1) !== "\xE2\x80\xA6"):
				$result['desc'] .= "\xE2\x80\xA6"; // "…"
			endif;
		endif;
		$snippet = esc ($result['desc']);
	else:
		$snippet = $url; // was the only option in Mouse 1.x
	endif;
	
	return "<li{$attributes}><a href='$url' title='$url'><span>$title</span><br><small>$snippet</small></a></li>";
}


/* Returns a $text marked up with HTML and CSS.
 */
function message (
	string $text
):	string
{
	return "<div style='background:yellow;color:#000;padding:1ex 1em'>"
		.  esc ($text)
		.  "</div>"
	;
}


/* Parses a comma-separated $string like 'foo=9, bar' and returns an array like
 * ['foo' => '9', 'bar' => 'bar']
 */
function parameters (
	string $string = NULL
):	array
{
	$array = [];
	if (isset ($string)):
		foreach (explode (',', $string) as $parameter):
			$p = explode ('=', $parameter, 2);
			$array[trim ($p[0])] = trim ($p[1] ?? $p[0]);
		endforeach;
	endif;
	return $array;
}


/* Normalizes a set of &$urls that define the realm of the search engine and
 * returns a short hash to identify that realm.
 */
function realm (
	array &$urls
):	string
{
	require_once __DIR__ . '/word.php';

	// Normalize to get a hash independent of things like lettercase of
	// case-insensitve URL components:
	foreach ($urls as &$url):
		$url = word::normalize_url ($url);
	endforeach;
	
	// Sort all but the first homepage to get a hash independent of the provided
	// order. The first one is special, since it serves as pars pro toto in
	// Atom feeds.
	if (count ($urls) > 2):
		$main = array_shift ($urls);
		sort ($urls, SORT_STRING | SORT_FLAG_CASE);
		array_unshift ($urls, $main);
	endif;

	// Return an 8-character long base-32 lowercase hash that avoids the chars
	// 0, 1 and l since the latter two can be mistaken for each other and 0 can
	// be mistaken for a big o by someone who does not know it is lowercase:
	$md5 = md5 ("MOUSE\t" . implode ("\t", $urls));
	return str_replace (['0', '1', 'l'], ['x', 'y', 'z'], base_convert (substr ($md5, 0, 5), 16, 32)
		.  str_pad (base_convert (substr ($md5, -5), 16, 32), 4, '0', STR_PAD_LEFT)
	);
}


/* Returns HTML script element that triggers the crawler.
 */
function trigger_crawler (
	string $module_url,
	string $id
):	string
{
	return "<script src='" . esc ($module_url . "/crawl.php?" . $id) . "' defer></script>";
}

// The End