<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2019-2022 Martin Janecke <mouse@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Prologue: Settings
 */

	// declare (strict_types = 1); // only during development

	namespace prlbr\mouse;

	// some time settings in seconds; further settings are in class retriever
	// for the maximum transfer time of a file and classes finder and feeder
	// for the maximum waiting time will the database is busy
	const MAXIMUM_EXECUTION_TIME = 28;
	const MINIMUM_CRAWLING_PAUSE = 29;
	const UPDATE_ROBOTSTXT_AFTER = 43200;

	// directories
	const PRIVATE_DIR = __DIR__ . '/../private/mouse';
	const PUBLIC_DIR  = __DIR__ . '/../public/mouse';

	// The script does not actually return Javascript code, but it can be
	// triggered by being referenced as <script src='…'></script> in HTML.
	header ('Content-Type: text/javascript');


/* Chapter 1: Set Error Handling
 *
 * We want all errors to be logged for debugging and set an error handler to do
 * so in Aneamal's cache directory and terminate execution afterwards.
 */

	// no type declarations here since they risk additional errors
	function log_error (
		$error   = 'Error',
		$level   = -1,
		$message = '',
		$file    = '',
		$line    = -1
	)
	{
		global $query;
		global $job;
		$query ??= '';
		$job ??= '';
		$log = PRIVATE_DIR . "/$query-fail-" . gmdate ('Ym') . '.txt';
		$msg = gmdate ('c') . ": $error/$level \"$message\" in file \"$file\", line $line, doing \"$job\".\n";
		@file_put_contents ($log, $msg, FILE_APPEND);
		exit;
	}
	
	error_reporting (E_ALL);

	set_error_handler (function ($level, $message, $file, $line) {
		error_reporting () & $level and log_error ('Error', $level, $message, $file, $line);
	});

	set_exception_handler (function ($e) {
		log_error ('Exception', $e->getCode (), $e->getMessage (), $e->getFile (), $e->getLine ());
	});


/* Chapter 2: The Task
 *
 * Which site shall be crawled can be read from a task file that is set up in
 * the Aneamal cache by index.php of this module. The task file's name contains
 * a hash which we expect to be given in the query string or as an argument when
 * this crawler is run via the command line. We check whether the task file
 * exists and compose the names of further files this crawler uses.
 * Foreseeable failures in this chapter are not written to the error log because
 * they could be triggered deliberately from the outside to spam the error log.
 * They are reported with leading // to make the a comment in Javascript, since
 * this crawler identifies as Javascript.
 */
 
	if (isset ($_SERVER['QUERY_STRING']) and $_SERVER['QUERY_STRING'] !== ''):
		$query = $_SERVER['QUERY_STRING'];
	elseif (isset ($_SERVER['argv'][1]) and $_SERVER['argv'][1] !== ''):
		$query = $_SERVER['argv'][1];
	else:
		exit ('// Failure: query string expected, but missing');
	endif;

	$vlength = strspn ($query, '23456789abcdefghijkmnopqrstuvxyz');
	if ($vlength !== strlen ($query)):
		exit ('// Failure: query string contains an unexpected character ' . ($query[$vlength] ?? ''));
	endif;

	$basename = PRIVATE_DIR . "/$query";
	$taskfile = "$basename/task.txt";
	$donefile = "$basename/done.txt";
	$newsfile = "$basename/news.txt";
	$jsonfile = "$basename/news.json";
	$nogofile = "$basename/nogo.txt";
	$dbnewfile = "$basename/" . strval (time ()) . '.db';
	$dbpattern = "$basename/*.db";

	if (!file_exists ($taskfile)):
		exit ('// Failure: task file missing for this query');
	endif;


/* Chapter 3: Cronjob Specials
 *
 * Caching by proxies or clients shall not last longer than the crawler pauses
 * between crawls, because it needs to be requested from us to do its work (it
 * crawls one page each time it runs). The requesting client does not need to
 * wait until the job is finished though, he just triggers it. So we close the
 * connection and keep the crawler running.
 */

	header ('Cache-Control: max-age=' . MINIMUM_CRAWLING_PAUSE . ', must-revalidate');
	header ('Expires: 0'); // fallback; "Cache-Control: max-age" has precedence

	ignore_user_abort (true);
	set_time_limit (MAXIMUM_EXECUTION_TIME);

	if (function_exists ('fastcgi_finish_request')):
		fastcgi_finish_request ();
	else:
		header ('Content-Length: 0');
		header ('Content-Encoding: none');
		header ('Connection: close');
		flush ();
	endif;


/* Chapter 4: Lock and Limit
 *
 * The done file stores which pages from the task file have already been
 * crawled, so it is written to every time a page is crawled. We use this to
 * implement a pause between crawls by checking when the file was last changed
 * and comparing that with the current time. To prevent two instances of the
 * crawler doing this at the same time and both finding that it is time to crawl
 * again, an exclusive file lock is also applied while checking the time.
 */

	// fails with an error, if the file is not writable/can not be created
 	$handle = fopen ($donefile, 'a');

	// exit silently, if the file is locked, otherwise lock it
 	if (!flock ($handle, LOCK_EX | LOCK_NB)):
	 	fclose ($handle);
	 	exit;
	endif;
 		
	// exit silently, if the last run is too short ago
	if (time () - filemtime ($donefile) < MINIMUM_CRAWLING_PAUSE):
		flock ($handle, LOCK_UN);
	 	fclose ($handle);
		exit;
	endif;

	// add a line break, which also updates the filemtime
	fwrite ($handle, "\n");
	
	// unlock; parallel work is still blocked for const MINIMUM_CRAWLING_PAUSE
	flock ($handle, LOCK_UN);
 	fclose ($handle);


/* Chapter 5: Dependencies
 *
 * Load classes to retrieve files over the internet and for adding content to
 * the SQLite database; define functions.
 */

	require_once __DIR__ . '/assets/feeder.php';
	require_once __DIR__ . '/assets/newsfeed.php';
	require_once __DIR__ . '/assets/retriever.php';

	/* Returns an array of URLs from $test which do no match the strings in
	 * $mask. An URL from $test matches a string in $mask, if the URL begins
	 * with the string.
	 */
	function drop_disallowed (
		array $test,
		array $mask
	):	array
	{
		foreach ($mask as $disallowed):
			$length = strlen ($disallowed);
			foreach ($test as $i => $url):
				if ($disallowed === substr ($url, 0, $length)):
					unset ($test[$i]);
				endif;
			endforeach;
		endforeach;
		return $test;
	}


/* Chapter 6: Jobs and Database Discovery
 *
 * Load the task file and done file; load disallowed patterns from robots.txt
 * files; determine from these which tasks are still to do. Also check which
 * databases exist and import a legacy database or create a new one, if none is
 * there yet.
 */

	if (!is_writable ($taskfile)):
		trigger_error ('task file not writable');
	endif;

	$task = file ($taskfile, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
	$init = explode ("\t", array_shift ($task));

	// get the list of URL patterns disallowed by robots.txt files and store it
	// in a file; create the file iff none exists yet
	if (file_exists ($nogofile) and time () - filemtime ($nogofile) < UPDATE_ROBOTSTXT_AFTER):
		$nogo = file ($nogofile, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
	else:
		$nogo = [];
		foreach ($init as $site):
			array_push ($nogo, ...retriever::fetch_robots_txt ($site));
		endforeach;
		$nogo = array_unique ($nogo);
		file_put_contents ($nogofile, implode ("\n", $nogo));
	endif;

	$done = file ($donefile, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);

	// We don't put disallowed pages on the task list, but what is disallowed
	// may have changed since then, so we do the check here as well.
	$todo = drop_disallowed (array_diff ($task, $done), $nogo);

	// get a list of databases; create one and try to import a legacy database,
	// iff none exists
	if (empty ($databases = glob ($dbpattern))):
		feeder::create ($databases[] = $dbnewfile);
		// try to import a legacy database and newsfile from Mouse <= 1.0.1
		$legacy = dirname (__DIR__) . "/cache/mouse-$query";
		if ($legacydbs = glob ("$legacy-*.db")):
			$legacydb = $legacydbs[max (0, count ($legacydbs) - 2)];
			copy ($legacydb, "$basename/" . substr ($legacydb, strrpos ($legacydb, '-') + 1));
			if (!file_exists ($newsfile) and is_readable ("$legacy-news.txt")):
				copy ("$legacy-news.txt", $newsfile);
				// create news feeds
				$feed = new newsfeed ($newsfile, [$legacydb]);
				$feed->save (PUBLIC_DIR, $query) or trigger_error ($feed->error);
				$feed->json_export ($jsonfile);
			endif;
		endif;
	endif;


/* Chapter 7: Optional Restart ("the crawler has finished, long may it crawl")
 *
 * If there are no jobs to do, the site has been indexed completely and we
 * prepare for a new round by resetting the task and done files and creating a
 * new database. The creation time is encoded in the database name. An old
 * database is pruned unless only few exist yet.
 */

	if (empty ($todo)):

		if (empty ($urls = drop_disallowed ($init, $nogo))):
			trigger_error ("'" . implode ("', '", $init) . "' blocked by robots.txt");
		endif;

		file_put_contents ($donefile, '');
		file_put_contents ($taskfile, implode ("\t", $init) . "\n" . implode ("\n", $urls));
		feeder::create ($dbnewfile);

		if (count ($databases) > 2):
			unlink (array_shift ($databases));
		endif;

		// create news feeds
		$feed = new newsfeed ($newsfile, [end ($databases)]);
		$feed->save (PUBLIC_DIR, $query) or trigger_error ($feed->error);
		$feed->json_export ($jsonfile);

		exit;
	endif;


/* Chapter 8: Database Access
 *
 * Connect to the newest available database. This is the one we will write to,
 * whereas the second newest is used to search in.
 */

	$db = new feeder (end ($databases));


/* Chapter 9: Claim a Job
 *
 * Everything is set up ... time to claim a job so that it won't be crawled next
 * time the crawler runs. It is not yet clear whether crawling it will succeed,
 * but that's intentional: if there's a bug triggered by the job, we want to
 * move on instead of automatically trying and failing again and again.
 */

	$job = array_shift ($todo);
	file_put_contents ($donefile, $job, FILE_APPEND);
	$done[] = $job;


/* Chapter 10: Fetch the Document
 *
 * Finally, the crawler retrieves a document and analyzes its content.
 */

	$doc = new retriever ($job, $init);


/* Chapter 11: Register New Jobs
 *
 * Links from the crawled file which are in scope of this search engine are
 * added to the task file unless they are already on it or disallowed by the
 * nogo list, i.e. excluded by a robots.txt. Links possibly include the redirect
 * location for HTTP 3xx status codes, which were not followed immediately when
 * the document was fetched.
 */

	$urls = array_diff ($doc->links, $task);

	if (!empty ($urls)):
		$urls = drop_disallowed ($urls, $nogo);
	endif;

	if (!empty ($urls)): // necessary again because $urls may have changed
		file_put_contents ($taskfile, "\n" . implode ("\n", $urls), FILE_APPEND);
	endif;


/* Chapter 12: Into the Database
 *
 * The document is registered in the database. Title and text are handled
 * separately, whereby for (X)HTML documents description and keywords are
 * handled together with the text.
 */

	if ($doc->indexable):
		$db->write (
			$job,
			$doc->title === ''? rawurldecode (basename ($job)): $doc->title,
			ltrim ($doc->description . ' ') . ltrim ($doc->keywords . ' ') . $doc->text,
			$doc->snippet,
			$doc->author
		);
	endif;
	unset ($db);


/* Chapter 13: Remember the First Time
 *
 * It is checked whether the document had been found before and, if not, is
 * added to the news file. This can be used for an atom feed for example.
 */

	if ($doc->indexable and $doc->title !== ''):
 		$news = [];
	 	if (file_exists ($newsfile)):
	 		foreach (file ($newsfile, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES) as $record):
	 			$news[] = substr ($record, 0, strcspn ($record, "\t"));
	 		endforeach;
	 	endif;
	 	if (!in_array ($job, $news)):
	 		file_put_contents ($newsfile, "\n$job\t" . strval (time ()), FILE_APPEND);
			// create news feeds
			$feed = new newsfeed ($newsfile, array_slice ($databases, -2));
			$feed->save (PUBLIC_DIR, $query) or trigger_error ($feed->error);
			$feed->json_export ($jsonfile);
	 	endif;
	 endif;


// The End